#include "profiler/timer.h"
#include "aliases.h"
#include <stdio.h>
#include <string.h>
#include <time.h>

#include <x86intrin.h>

typedef struct {
  profiler_sample_t samples[MAX_PROFILE_SAMPLES];
  u64 cpu_freq;
  u64 size;
  u64 start;
  u64 end;
  u64 max_title_length;
  profiler_sample_t *active;
} profiler_t;

INTERNAL profiler_t profiler = {0};

typedef struct timespec timespec_t;

u64 get_os_frequency() { return 1000000000; }

u64 get_os_time(void) {
  timespec_t ts = {0};

  if (clock_gettime(CLOCK_MONOTONIC_RAW, &ts) != 0) {
    return 0;
  }

  return ts.tv_sec * get_os_frequency() + ts.tv_nsec;
}

u64 read_cpu_timer(void) { return __rdtsc(); }

u64 get_cpu_freq(u64 milliseconds) {
  u64 os_freq = get_os_frequency();
  u64 os_start = get_os_time();
  u64 cpu_start = read_cpu_timer();
  u64 os_end = 0;
  u64 os_elapsed = 0;
  u64 os_wait_time = os_freq * milliseconds / 1000;

  while (os_elapsed < os_wait_time) {
    os_end = get_os_time();
    os_elapsed = os_end - os_start;
  }

  u64 cpu_end = read_cpu_timer();
  u64 cpu_elapsed = cpu_end - cpu_start;

  u64 cpu_freq = 0;

  if (os_elapsed) {
    cpu_freq = cpu_elapsed * os_freq / os_elapsed;
  }

  return cpu_freq;
}

void profile_start(u64 count) {
  profiler.cpu_freq = get_cpu_freq(500);
  profiler.start = read_cpu_timer();
  profiler.max_title_length = 0;
  profiler.size = count;
}

void profile_end() {
  if (!profiler.start || !profiler.size) {
    return;
  }

  profiler.end = read_cpu_timer();

  u64 total = 0;

  if (profiler.end >= profiler.start) {
    total = profiler.end - profiler.start;
  }

  u16 time_precision = 16;
  u16 time_char_count = 20;
  u16 duration_char_count = 22;
  u16 hits_char_count = 10;
  u16 percentage_precision = 8;
  u16 percentage_char_count = 12;

  // clang-format off
  printf("\n==============================PROFILING==============================\n");
  // clang-format on

  if (profiler.cpu_freq) {
    printf("Total: %*.*f seconds (CPU frequency: %llu hz/sec)\n\n",
           time_char_count, time_precision, (f64)total / profiler.cpu_freq,
           (unsigned long long)profiler.cpu_freq);
  }

  profiler_sample_t *sample = NULL;

  for (u64 i = 0; i < profiler.size; ++i) {
    sample = &(profiler.samples[i]);

    if (sample->hit_count == 0) {
      continue;
    }

    u64 duration = 0;

    if (sample->duration >= sample->children_duration) {
      duration = sample->duration - sample->children_duration;
    }

    printf("%*s (hits: %*lld): %*lld (%*.*f %%", (i32)profiler.max_title_length,
           sample->title, hits_char_count,
           (unsigned long long)sample->hit_count, duration_char_count,
           (unsigned long long)duration, percentage_char_count,
           percentage_precision, (f64)duration / total * 100.0);

    if (sample->children_duration > 0) {
      printf(", w/ children: %*.*f %%", percentage_char_count,
             percentage_precision, (f64)sample->duration / total * 100.0);
    }

    printf(")\n");
  }
}

void sample_start(u64 id, const char *title) {
  if (id >= MAX_PROFILE_SAMPLES) {
    return;
  }

  profiler_sample_t *sample = &(profiler.samples[id]);

  if (!(sample->title) || strcmp(title, sample->title) != 0) {
    sample->title = title;
    sample->start = 0;
    sample->duration = 0;
    sample->children_duration = 0;
    sample->hit_count = 0;
    sample->parent = NULL;

    u64 length = strlen(sample->title);

    if (length > profiler.max_title_length) {
      profiler.max_title_length = length;
    }
  }

  sample->start = read_cpu_timer();
  ++(sample->hit_count);

  if (sample != profiler.active) {
    // This handles recursive functions by changing the parent only when a
    // function isn't calling itself
    sample->parent = profiler.active;
  }

  profiler.active = sample;
}

void sample_end(u64 id) {
  if (id >= MAX_PROFILE_SAMPLES) {
    return;
  }

  profiler_sample_t *sample = &(profiler.samples[id]);

  u64 duration = read_cpu_timer() - sample->start;

  if (!(sample->parent) || sample != sample->parent) {
    // This handles recursive functions by adding to the duration only when a
    // function isn't calling itself
    sample->duration += duration;
  }

  if (sample->parent && sample != sample->parent) {
    // This handles recursive functions by adding to the children_duration only
    // when a function isn't calling itself
    sample->parent->children_duration += duration;
  }

  profiler.active = sample->parent;
}
